/* xbutton.c 
 *
 * Copyright (C) 2000 Transmeta Corporation
 *
 * written by Nathan Laredo <nlaredo@transmeta.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/Xmd.h>
#include <X11/extensions/XTest.h>
#include <X11/extensions/xtestext1.h>

#undef XKEYBOARD_SIGNAL_INTERFACE

//#define DEBUG_BY_HITACHI
#undef DEBUG_BY_HITACHI
#define INIT_FILE "/home/user/.xbuttonrc"
#define MAX_PATH_LEN 255

void ExecProcess(char *pszExec)
{
	pid_t child = 0;

	//if argument is blank, return
	if(!strlen(pszExec)){
		return;
	}

	child = fork();
	switch(child){
	case 0:		//Child
		child = fork();
		switch(child){
		case 0:		//Child-Child
			execlp(pszExec, pszExec, NULL);
		case -1:	//Child-Error
			fprintf(stderr, "fork in frok error.\n");
			exit(-1);
		default:	//Child
			exit(0);
		}
	case -1:	//Error
		fprintf(stderr, "fork error.\n");
		return;
	default:	//Parent(Own)
		waitpid(child, NULL, 0);
	}
	return;
}

int main(int argc, char **argv)
{
	int fd, myscreen, realkey = 0, lasttime = 500000;
	Display *mydisplay;
	fd_set rdfs;
	struct timeval tv;
	unsigned char buttonstate = 0;
	unsigned char buttonexecute = 0;
        FILE *load;
        int len, icount;
	char szExecutePath[3][MAX_PATH_LEN];	//0:Mail, 1:Internet, 2:Program
	int kri_max=500000, kri_min=100;	//key Repeat Interval(microsec)

#ifdef XKEYBOARD_SIGNAL_INTERFACE
	int key_pid;

	if (argc < 2) {
		fprintf(stderr, "usage: %s xkeyboard_pid\n", argv[0]);
		exit(1);
	}
	if ((key_pid = atoi(argv[1])) < 2 || kill(key_pid, 0) < 0) {
		fprintf(stderr, "please specify a valid pid for xkeyboard\n");
		exit(1);
	}
#endif
	if ((fd = open("/dev/tmbutton", O_RDONLY)) < 0) {
		perror("open /dev/tmbutton");
		exit(1);
	}
	if (!getenv("DISPLAY")) {
		fprintf(stderr, "DISPLAY not set.\n");
		exit(1);
	}
	if (!(mydisplay = XOpenDisplay(getenv("DISPLAY")))) {
		fprintf(stderr, "Unable to open display.\n");
		exit(1);
	}
	myscreen = DefaultScreen(mydisplay);
	
	//Read InitFile for init value
	for (icount = 0; icount < 3; icount++){
		szExecutePath[icount][0] = '\0';
	}
	kri_max=500000; kri_min=100;

	if ((load = fopen(INIT_FILE, "r")) != NULL) {
		for (icount = 0; icount < 3; icount++){
			len = fscanf(load, "%s\n", szExecutePath[icount]);
			if (len != 1){
				szExecutePath[icount][0] = '\0';
				break;
			}
		}
		len = fscanf(load, "%d %d\n", &kri_max, &kri_min);
		if (len != 2){
			kri_max=500000; kri_min=100;
		}
		fclose(load);
	}
	lasttime = kri_max;
#ifdef DEBUG_BY_HITACHI
	printf("Mail Execute File Path     = %s.\n", szExecutePath[0]);
	printf("Internet Execute File Path = %s.\n", szExecutePath[1]);
	printf("Program Execute File Path  = %s.\n", szExecutePath[2]);
	printf("Key Repeat Intervall(micro sec) MAX=[%d] MIN=[%d].\n", 
		kri_max, kri_min);
#endif
 
	while (1) {
		realkey = 0;
		/* key repeat timeout */
		tv.tv_sec = 0;			// seconds
		tv.tv_usec = lasttime;		// microseconds
		FD_ZERO(&rdfs);
		FD_SET(fd, &rdfs);
		//wait for button push or release(select function)
		if (!buttonstate ||	/* if no keys to repeat, block */
		    select(FD_SETSIZE, &rdfs, NULL, NULL, &tv)) {
			//two case:
			//	1)no button push(interrupt) before
			//	2)interrupt happen
			if (read(fd, &buttonstate, 1) < 1)
				break;
			realkey++;
			lasttime = kri_max;
#ifdef DEBUG_BY_HITACHI
	printf("read buttonstate = %x.\n", buttonstate);
#endif
		}	//if no interrupt, do nothing
		// Button push -> interrupt happen -> new buttonstate(!=0)
		// Button pushing -> no interrupt -> maintain buttonstate
		// Button release -> interrupt happen -> new buttonstate(==0)
		// written by T.Arai in HITACHI on 2000.10.25
		// update by T.Arai in HITACHI on 2001.02.14

		if (buttonstate & 0x01)	{	/* up */
			XTestFakeKeyEvent(mydisplay, 98, 1, 0);
			XTestFakeKeyEvent(mydisplay, 98, 0, 0);
		}
		if (buttonstate & 0x02) {	/* down */
			XTestFakeKeyEvent(mydisplay, 104, 1, 0);
			XTestFakeKeyEvent(mydisplay, 104, 0, 0);
		}
		if (buttonstate & 0x04) {	/* left */
			XTestFakeKeyEvent(mydisplay, 100, 1, 0);
			XTestFakeKeyEvent(mydisplay, 100, 0, 0);
		}
		if (buttonstate & 0x08) {	/* right */
			XTestFakeKeyEvent(mydisplay, 102, 1, 0);
			XTestFakeKeyEvent(mydisplay, 102, 0, 0);
		}
		if (buttonstate & 0x10) {	/* enter */
			XTestFakeKeyEvent(mydisplay, 36, 1, 0);
			XTestFakeKeyEvent(mydisplay, 36, 0, 0);
		}
		if (buttonstate & 0x20) {	/* mail Button */
			if (!(buttonexecute & 0x20)){
				ExecProcess(szExecutePath[0]);
				buttonexecute |= 0x20;
			}		
		}else if (buttonexecute & 0x20){
			buttonexecute &= ~0x20;
		}
		if (buttonstate & 0x40) {	/* Internet Button */
			if (!(buttonexecute & 0x40)){
				ExecProcess(szExecutePath[1]);
				buttonexecute |= 0x40;
			}		
		}else if (buttonexecute & 0x40){
			buttonexecute &= ~0x40;
		}
		if (buttonstate & 0x80) {	/* Program Button */
			if (!(buttonexecute & 0x80)){
				ExecProcess(szExecutePath[2]);
				buttonexecute |= 0x80;
			}		
		}else if (buttonexecute & 0x80){
			buttonexecute &= ~0x80;
		}
		goto realkey_skip;
		if (realkey && buttonstate & 0x20) {	/* keyboard button */
#ifdef XKEYBOARD_SIGNAL_INTERFACE
			kill(key_pid, SIGUSR1);	/* toggle xkeyboard state */
#else
			XTestFakeMotionEvent(mydisplay, myscreen, 10,
				DisplayHeight(mydisplay, myscreen) - 200, 0);
			XTestFakeButtonEvent(mydisplay, 1, 1, 0);
			XTestFakeButtonEvent(mydisplay, 1, 0, 0);
#endif
		}
realkey_skip:
		// if continue to push key, repeat more fast 
		if (buttonstate) {
			XFlush(mydisplay);
			lasttime >>= 1;
			if (lasttime < kri_min)
				lasttime = kri_min;
		}
	}
	XCloseDisplay(mydisplay);
	exit(0);
}
